import os,subprocess,re,sys,shlex
from subprocess import PIPE

def extractsub(mkv,keywords=['default'],language='eng',ffmpeg = 'ffmpeg'):
    r, ext = os.path.splitext(mkv)
    srt = r+'.srt'
    cmd = shlex.split(f'''{ffmpeg} -hide_banner -i "{mkv}"''')
    result = subprocess.run(cmd,stdout=PIPE, stderr=PIPE, universal_newlines=True)
    # print(result.returncode, result.stdout, result.stderr)
    streams = {}
    print(' '.join(cmd))
    n=-1
    streamid = None
    for line in result.stderr.split('\n'):
        if n==1:
            m = re.match(' +title *: (.*)',line)
            if m:
                streams.setdefault(streamid,dict())['title'] = str(m.group(1))
            m = re.match(' +BPS.* *: (.*)',line)
            if m:
                streams.setdefault(streamid,dict())['BPS'] = str(m.group(1))
            m = re.match(' +DURATION.* *: (.*)',line)
            if m:
                streams.setdefault(streamid,dict())['DURATION'] = str(m.group(1))
            m = re.match(' +NUMBER_OF_FRAMES.* *: (.*)',line)
            if m:
                streams.setdefault(streamid,dict())['NUMBER_OF_FRAMES'] = str(m.group(1))
            m = re.match(' +NUMBER_OF_BYTES.* *: (.*)',line)
            if m:
                streams.setdefault(streamid,dict())['NUMBER_OF_BYTES'] = str(m.group(1))
            m = re.match(' +_STATISTICS_WRITING_APP.* *: (.*)',line)
            if m:
                streams.setdefault(streamid,dict())['_STATISTICS_WRITING_APP'] = str(m.group(1))
            m = re.match(' +_STATISTICS_WRITING_DATE_UTC.* *: (.*)',line)
            if m:
                streams.setdefault(streamid,dict())['_STATISTICS_WRITING_DATE_UTC'] = str(m.group(1))
            m = re.match(' +_STATISTICS_TAGS.* *: (.*)',line)
            if m:
                streams.setdefault(streamid,dict())['_STATISTICS_TAGS'] = str(m.group(1))
        if n==0:
            m = re.match(' +Metadata:',line)
            if m:
                n=1
        if n==1 or n==-1:
            m = re.match(' +Stream #(\d+:\d+)\((\w+)\): (\w+): (.*)',line)
            if m:
                n=0
                streamid = str(m.group(1))
                streams.setdefault(streamid,dict())['language'] = str(m.group(2))
                streams.setdefault(streamid,dict())['type'] = str(m.group(3))
                streams.setdefault(streamid,dict())['subtype'] = str(m.group(4))
    selectstream = None
    s={}
    print(f'{language} streams:')
    findlang = False
    otherlangs = []
    for streamid,v in streams.items():
        if v['type'] == 'Subtitle':
            if v['language'] == language:
                findlang = True
                if 'hdmv_pgs_subtitle' in v['subtype']:
                    continue
                fmt = None
                if 'subrip' in v['subtype']:
                    fmt = '.srt'
                if 'ass' in v['subtype']:
                    fmt = '.ass'
                if fmt == None:
                    print(f'Unknow format: {streamid} {mkv}')
                for keyword in keywords:
                    if keyword in v['subtype'] or keyword in v.get('title',''):
                        # select by first keywords
                        selectstream = [streamid,fmt]
                        break
                else:
                    s[streamid] = v
                print('  ',streamid,v['subtype'],v.get('title',''),v.get('NUMBER_OF_BYTES','-1'))
            else:
                otherlangs.append(f'  {v["language"]} {streamid}: {mkv}')
    if not findlang:
        print(f'Error: language({language}) not found:\n','\n'.join(otherlangs))
    if not selectstream or True:
        # select by min size
        lastsize = 0
        for streamid,v in s.items():
            size = int((v.get('NUMBER_OF_BYTES','0')).strip())
            if size < lastsize or lastsize==0:
                if 'subrip' in v['subtype']:
                    fmt = '.srt'
                if 'ass' in v['subtype']:
                    fmt = '.ass'
                selectstream = [streamid,fmt]
                lastsize = size
    res = ''
    if selectstream:
        if selectstream[1] == None:
            print(f'Error: Unknow format: {selectstream[0]} {mkv}')
        else:
            out = r + selectstream[1]
            print(f"Extract: {selectstream[0]} in {mkv}")
            subprocess.run(shlex.split(f'''{ffmpeg} -v error -stats -i "{mkv}" -vn -an -map {selectstream[0]} "{out}"'''))
    else:
        for streamid,v in s.items():
            fmt = '.None'
            if 'subrip' in v['subtype']:
                fmt = '.srt'
            if 'ass' in v['subtype']:
                fmt = '.ass'
            # not select stream, print commands
            res += f'''Error: Not extract: {ffmpeg} -v error -stats -i "{mkv}" -vn -an -map {streamid} "{r + fmt}" \n'''
    print()
    return res

def run(mkv):
    # res = extractsub(mkv,keywords=['default'],language='eng',ffmpeg='ffmpeg.exe')
    res = extractsub(mkv,keywords=['default'],language='eng',ffmpeg='ffmpeg')
    return res

if __name__ == '__main__':
    if len(sys.argv)>1:
        failedlist = []
        for file in sys.argv[1:]:
            f = os.path.abspath(file)   
            if os.path.exists(f):
                if os.path.isdir(f):
                    for root, dirs, files in os.walk(os.path.abspath(f)):
                        for name in files:
                            mkv = os.path.join(root, name)
                            r, ext = os.path.splitext(mkv)
                            if ext =='.mkv':
                                res = run(mkv)
                                if len(res)>0:
                                    failedlist.append(mkv)
                                    print(res)
                elif os.path.isfile(f):
                    mkv = f
                    r, ext = os.path.splitext(mkv)
                    if ext =='.mkv':
                        res = run(mkv)
                        if len(res)>0:
                            failedlist.append(mkv)
                            print(res)
            else:
                print(f'\nError: file or folder not exist: {f}')
        if len(failedlist)>0:
            print(f'\nFail to extract:\n'+'\n'.join(failedlist))
    else:
        print("Usage: python extractsub.py 1.mkv 2.mkv folder1 folder2 ...")
        